# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
from typing import Tuple

separator:str = os.sep


def splitext(path:str) -> Tuple[str, str]:
    """A simpler and faster version of os.path.splitext().
    """
    sep:int = path.rfind(separator)
    dot:int = path.rfind(".")
    if dot > sep:
        idx:int = sep + 1
        while idx < dot:
            if path[idx:idx+1] != ".":
                return path[:dot], path[dot:]
            idx += 1

    return path, path[:0]

def split(path:str) -> Tuple[str, str]:
    """A simpler and faster version of os.path.split().
    """
    sep:int = path.rfind(separator)
    if sep < 0:
        return "", path
    else:
        return path[:sep], path[sep + 1:]

def join(part1:str, part2:str) -> str:
    """A simpler and faster version of os.path.join().
    """
    if part1:
        if part2:
            if part1 == separator:
                return part1 + part2
            else:
                return part1 + separator + part2
        else:
            return part1
    else:
        return part2



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import mimetypes

from libff.plugin import *


class Medium(Plugin):
    """The "medium" plugin provides access to information from media files like image, audio and
       video files, e.g. image format, running time, mp3 tags, etc. It requires the 'pymediainfo'
       module.
    """

    speed = Speed.SLOW

    use_cache = True
    cache_tag = 0

    attributes = [
        ("duration", Duration, "The duration of a medium (audio, video) in seconds."),
        ("artist", String, "The artist audio tag of the file."),
        ("album", String, "The album audio tag of the file."),
        ("title", String, "The title audio tag of the file."),
        ("genre", String, "The genre audio tag of the file."),
        ("date", String, "The date audio tag of the file."),
        ("format", String, "The format of an image ('png', 'jpeg', etc.) in case the file is an "\
                           "image."),
        ("width", Number, "The width of a visual medium (image, video) in pixel."),
        ("height", Number, "The height of a visual medium (image, video) in pixel.")
    ]

    @classmethod
    def setup(cls):
        # pylint:disable=global-statement,import-outside-toplevel
        global pymediainfo
        try:
            import pymediainfo
        except ImportError:
            raise MissingImport("pymediainfo")

    def prepare_tracks(self, entry):
        """Prepare a more accessible structure of tracks from a media file.
        """
        # pylint:disable=undefined-variable
        try:
            tracks = {}
            for track in pymediainfo.MediaInfo.parse(entry.path,
                    encoding_errors="backslashreplace").tracks:
                track_type = track.track_type.lower()
                if track_type in ("audio", "video", "image") and track_type not in tracks:
                    tracks[track_type] = track
                elif track_type == "general":
                    tracks.setdefault(track_type, []).append(track)
            return tracks

        except (OSError, RuntimeError):
            raise NoData

    def can_handle(self, entry):
        mimetype = mimetypes.guess_type(entry.name)[0]
        if not mimetype:
            return False
        return mimetype.split("/", 1)[0] in set(["image", "audio", "video"])

    def cache(self, entry):
        cached = {}

        tracks = self.prepare_tracks(entry)

        audio_track = tracks.get("audio")
        video_track = tracks.get("video")
        image_track = tracks.get("image")

        # Extract available audio tags.
        for track in tracks.get("general", []):
            for key, attr in (("artist", "performer"), ("album", "album"), ("title", "title"),
                    ("genre", "genre"), ("date", "recorded_date")):
                value = getattr(track, attr, "")
                if value:
                    cached[key] = value

        # Extract the width and height of a video or an image. If there are both video and image
        # streams give precedence to video.
        if video_track is not None:
            if video_track.height is not None and video_track.width is not None:
                cached["width"] = video_track.width
                cached["height"] = video_track.height

        elif image_track is not None:
            if image_track.format is not None:
                cached["format"] = image_track.format.lower()
            if image_track.height is not None and image_track.width is not None:
                cached["width"] = image_track.width
                cached["height"] = image_track.height

        # Extract the duration of an audio or video stream.
        for track in (video_track, audio_track):
            if track is None:
                continue

            duration = track.duration
            if duration is not None:
                if isinstance(duration, str):
                    duration = float(duration)
                cached["duration"] = int(duration / 1000)

        return cached

    def process(self, entry, cached):
        return cached



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import shlex
import argparse

from .main import _Base
from .arguments import Defaults, ArgumentsPostProcessor
from .processing import ImmediateGenerator, CollectiveGenerator


class FindNamespace(argparse.Namespace):
    """An argparser.Namespace that contains the minimum number of arguments for setting up the
       Python API.
    """
    # pylint:disable=too-many-instance-attributes

    debug = ["none"]
    profile = None
    json = None

    depth = None

    # Options that have no relevance for the Python API but will be referenced somewhere in the
    # code. Would be nice to remove these one day.
    count = None
    exec = None
    exec_batch = None
    action = None
    help = None

    no_parent_ignore = False


class Find(_Base):
    """Query the filesystem and yield the resulting entries.
    """
    # pylint:disable=too-many-arguments,too-many-locals,dangerous-default-value,invalid-name

    def __init__(self, query=None, directories=["."], exclude=[], output=["file"],
            sort=[], reverse=False,
            hide=False, ignore=False, one_file_system=False,
            case=Defaults.case,
            si=Defaults.si,
            follow_symlinks=Defaults.follow_symlinks,
            absolute_path=False, jobs=Defaults.jobs, cache=Defaults.cache):
        super().__init__()

        if case not in Defaults.case_choices:
            raise ValueError(f"case argument must be one of {', '.join(Defaults.case_choices)}")

        if query is None:
            tests = []
        elif isinstance(query, str):
            tests = shlex.split(query)
        else:
            tests = query

        args = FindNamespace()
        args.tests = tests
        args.directories = directories.copy()
        args.exclude = exclude.copy()
        args.output = output
        args.sort = sort
        args.reverse = reverse
        args.hide = hide
        args.ignore = ignore
        args.one_file_system = one_file_system
        args.case = case
        args.si = si
        args.follow_symlinks = follow_symlinks
        args.absolute_path = absolute_path
        args.jobs = jobs
        args.cache = cache

        processor = ArgumentsPostProcessor(args)
        processor.process()

        self.setup_context(args, [])
        self.setup_components()
        self.setup_processing()
        if __debug__:
            self.show_debug_info()
        self.setup_walker()

    def setup_processing(self):
        if self.context.args.sort:
            self.context.processing = CollectiveGenerator(self.context)
        else:
            self.context.processing = ImmediateGenerator(self.context)

    def __iter__(self):
        # Start processing.
        try:
            self.context.walker.start_processes()

            # Collect Entry objects from the Walker.
            self.context.processing.loop()

            # Finalize and put out the result if --sort or --exec-batch is involved.
            self.context.processing.finalize()

            yield from self.context.processing

        except KeyboardInterrupt:
            # Stop all processes immediately.
            self.context.stop()
            raise

        self.context.close()
        self.context.walker.close()
        self.context.processing.close()



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import grp
import pwd
import stat

from .path import join, split, splitext
from .type import Mode


class EntryAttributeError(AttributeError):
    """A subclass of AttributeError that provides the missing attribute name.
    """

    def __init__(self, name):
        self.name = name
        super().__init__(name)


class StartDirectory:
    """A basic version of an Entry that is used for start directories and is passed to Entry
       objects as the reference point for relpath, depth and samedev.
    """

    def __init__(self, args, root):
        self.root = root
        self.absroot = os.path.abspath(root)
        self.status = os.stat(self.root, follow_symlinks=args.follow_symlinks)
        self.device = self.status.st_dev


# pylint:disable=too-many-instance-attributes,too-many-public-methods
class Entry:
    """Provide access to all of a file's information. The attributes and properties have to adhere
       to the attributes dict of the File plugin, because it just redirects requests right to the
       Entry object.
    """

    _pwd_cache = {}
    _grp_cache = {}

    @classmethod
    def as_reference(cls, args, path):
        """Create an Entry from scratch, e.g when a file is given as reference. The resulting Entry
           object has emulated values for depth and samedev.
        """
        # This is a bit kludgy, but here we try to attribute a reference file to a search directory
        # from the command line, in order to emulate attributes like depth and samedev. If the
        # reference file is outside of the search area we use default values.
        abspath = os.path.abspath(path)
        for directory in args.directories:
            # Go through the list of search directories and pick the first one that matches the
            # reference file.
            dirname = os.path.abspath(directory) + os.sep
            if abspath.startswith(dirname):
                relpath = abspath[len(dirname):]
                break
        else:
            directory, relpath = split(path)

        start_directory = StartDirectory(args, directory)
        status = os.stat(path, follow_symlinks=args.follow_symlinks)

        return cls(start_directory, relpath, status)

    def __init__(self, start_directory, relpath, status, ignore_paths=None):
        # pylint:disable=too-many-branches

        self.start_directory = start_directory
        self.relpath = relpath
        self.status = status
        self.ignore_paths = ignore_paths

        self.root = self.start_directory.root
        self.path = join(self.root, self.relpath) if self.root != "." else self.relpath
        self.abspath = join(self.start_directory.absroot, self.relpath)

        self.dir, self.name = split(self.path)

        # Collect information early to avoid having OSErrors later on.
        if stat.S_ISLNK(self.status.st_mode):
            self.link = os.readlink(self.path)
            self.target = os.path.realpath(join(self.dir, self.link))
            self.broken = not os.path.exists(self.target)
        else:
            # Don't set link and target so that KeyError is raised in get_attribute().
            self.broken = False

        self.hide = self.name[0] == "."

        self.mode = self.status.st_mode

        if stat.S_ISDIR(self.mode):
            self.type = "directory"
        elif stat.S_ISREG(self.mode):
            self.type = "file"
        elif stat.S_ISLNK(self.mode):
            self.type = "symlink"
        elif stat.S_ISSOCK(self.mode):
            self.type = "socket"
        elif stat.S_ISFIFO(self.mode):
            self.type = "fifo"
        elif stat.S_ISCHR(self.mode):
            self.type = "char"
        elif stat.S_ISBLK(self.mode):
            self.type = "block"
        elif stat.S_ISDOOR(self.mode):
            self.type = "door"
        elif stat.S_ISPORT(self.mode):
            self.type = "port"
        elif stat.S_ISWHT(self.mode):
            self.type = "whiteout"
        else:
            self.type = "other"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.path!r}>"

    #
    # Public properties.
    #
    @property
    def ext(self):
        """The file extension without the leading dot or the empty string if the file has no
           extension.
        """
        return splitext(self.path)[1][1:]

    @property
    def device(self):
        """The number of the device the file is located.
        """
        return self.status.st_dev

    @property
    def inode(self):
        """The number of the inode of the file.
        """
        return self.status.st_ino

    @property
    def samedev(self):
        """Whether the file is on the same device as the start directory.
        """
        return self.device == self.start_directory.device

    @property
    def depth(self):
        """The depth of the file relative to the start directory.
        """
        return self.relpath.count(os.sep)

    @property
    def exec(self):
        """Whether the file is executable or not.
        """
        return self.is_executable()

    @property
    def size(self):
        """The size of the file in bytes. All types except 'file' have a size of 0.
        """
        if stat.S_ISREG(self.status.st_mode):
            return self.status.st_size
        else:
            return 0

    @property
    def time(self):
        """An alias for 'mtime'.
        """
        return self.mtime

    @property
    def mtime(self):
        """The modification time of the file in seconds since epoch.
        """
        return int(self.status.st_mtime)

    @property
    def ctime(self):
        """The inode change time of the file in seconds since epoch.
        """
        return int(self.status.st_ctime)

    @property
    def atime(self):
        """The access time of the file in seconds since epoch.
        """
        return int(self.status.st_atime)

    @property
    def perm(self):
        """The permission bits of the file without the file type bits.
        """
        return self.status.st_mode & Mode.MODE_ALL

    @property
    def links(self):
        """The number of links to the inode.
        """
        return self.status.st_nlink

    @property
    def uid(self):
        """The user id of the owner of the file.
        """
        return self.status.st_uid

    @property
    def gid(self):
        """The group id of the owner of the file.
        """
        return self.status.st_gid

    @property
    def user(self):
        """The user name of the owner of the file.
        """
        try:
            return self._pwd_cache[self.status.st_uid]
        except KeyError:
            uid = self.status.st_uid
            try:
                self._pwd_cache[uid] = pwd.getpwuid(uid)[0]
            except KeyError:
                self._pwd_cache[uid] = ""
            return self._pwd_cache[uid]

    @property
    def group(self):
        """The group name of the owner of the file.
        """
        try:
            return self._grp_cache[self.status.st_gid]
        except KeyError:
            gid = self.status.st_gid
            try:
                self._grp_cache[gid] = grp.getgrgid(gid)[0]
            except KeyError:
                self._grp_cache[gid] = ""
            return self._grp_cache[gid]

    @property
    def hidden(self):
        """Whether the file is "hidden" or not, i.e. if one of the path components contains a
           leading dot.
        """
        return any(part.startswith(".") for part in self.path.split(os.sep))

    @property
    def empty(self):
        """Whether the file or directory is empty or not.
        """
        if stat.S_ISDIR(self.status.st_mode):
            # The far better solution would be to set some kind of "empty" attribute in
            # FilesystemWalker, so that we don't have to call os.scandir() two times for every
            # directory. But the way FilesystemWalker is designed makes this impossible: the
            # directory has already been yielded before we find out if it is empty or not.
            with os.scandir(self.path) as entries:
                for _ in entries:
                    return False
                return True
        elif stat.S_ISREG(self.status.st_mode):
            return self.size == 0
        else:
            return False

    @property
    def pathx(self):
        """The file path without the extension.
        """
        return splitext(self.path)[0]

    @property
    def namex(self):
        """The file basename without the extension.
        """
        return os.path.basename(self.pathx)

    @property
    def mount(self):
        """Whether the entry is a mountpoint.
        """
        if not self.is_dir():
            return False

        parent = os.path.realpath(join(self.path, ".."))

        if parent == self.path:
            return True

        try:
            status = os.lstat(parent)
        except OSError:
            return False

        return self.status.st_dev != status.st_dev

    #
    # Private properties.
    #
    @property
    def dirname(self):
        """The dirname for the path.
        """
        return os.path.abspath(self.dir)

    @property
    def basename(self):
        """The basename for the path.
        """
        return self.name

    def is_dir(self):
        """Return True if the Entry is a directory.
        """
        return self.type == "directory"

    def is_file(self):
        """Return True if the Entry is a regular file.
        """
        return self.type == "file"

    def is_symlink(self):
        """Return True if the Entry is a symbolic link.
        """
        return self.type == "symlink"

    def is_socket(self):
        """Return True if the Entry is a socket.
        """
        return self.type == "socket"

    def is_fifo(self):
        """Return True if the Entry is a fifo i.e. named pipe.
        """
        return self.type == "fifo"

    def is_executable(self):
        """Return True if the Entry is executable.
        """
        if self.is_dir() or self.is_symlink():
            return False
        return ((self.perm & stat.S_IXUSR) | (self.perm & stat.S_IXGRP) |
                (self.perm & stat.S_IXOTH)) > 0

    def __getattr__(self, name):
        raise EntryAttributeError(name)

    def get_attribute(self, name):
        """Return attribute `name` from this Entry object and raise KeyError if the attribute is
           not defined.
        """
        try:
            return getattr(self, name)
        except EntryAttributeError as exc:
            # We have to check if the AttributeError that was raised was actually about
            # attribute.name, so we don't hide programming errors inside the Entry object.
            if exc.name != name:
                raise AttributeError(exc.name)

            raise KeyError(name)



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from libff.plugin import *


class Grep(Plugin):
    """The "grep" plugin provides access to the lines and the line count of a text file. It is
       useful for finding files that contain certain lines or certain substrings.
    """

    attributes = [
        ("text",    Boolean, "Whether the file contains text or binary data."),
        ("linecount", Number, "The number of lines in the file."),
        ("lines", ListOfStrings, "The lines of the file.")
    ]

    def can_handle(self, entry):
        return entry.is_file()

    def process(self, entry, cached):
        try:
            with open(entry.path) as fobj:
                try:
                    lines = fobj.readlines()
                except UnicodeDecodeError:
                    yield "text", False
                else:
                    yield "text", True
                    yield "linecount", len(lines)
                    yield "lines", lines
        except OSError:
            raise NoData



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import collections


class _Attribute(collections.namedtuple("Attribute", "plugin name")):
    """An Attribute namedtuple that contains the plugin name and the attribute name.
    """

    def __new__(cls, plugin:str, name:str, fullname:str, alias:str):
        self = super().__new__(cls, plugin, name)
        self.fullname:str = fullname
        self.alias:str = alias if alias is not None else fullname
        return self

    def __str__(self):
        # pylint:disable=no-member
        return self.alias

def Attribute(plugin:str, name:str, alias:str=None) -> _Attribute:
    """Factory function to create Attribute objects.
    """
    # pylint:disable=invalid-name
    return _Attribute(plugin, name, f"{plugin}.{name}", alias)



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import tarfile

from libff.plugin import *


class Tar(Plugin):
    """The "tar" plugin provides information about tar archives. It is useful for finding file
       names inside tar archives. It currently uses the tarfile module from the Python Standard
       Library which limits the supported compression methods to gzip(1), bzip2(1), xz(1) and
       uncompressed.
    """

    speed = Speed.SLOW

    use_cache = True

    attributes = [
        ("members", ListOfStrings, "The list of file names that are stored in a .tar archive.")
    ]

    extensions = (
            ".tar",
            ".tar.gz", ".taz", ".tgz", ".tpz",
            ".tar.bz2", ".tb2", ".tbz", ".tbz2", ".tz2",
            ".tar.lz", ".tar.lzma", ".tlz",
            ".tar.xz", ".txz")

    def can_handle(self, entry):
        return entry.name.endswith(self.extensions)

    def cache(self, entry):
        try:
            with tarfile.open(entry.path) as tar:
                return set(name for name in tar.getnames())
        except (OSError, EOFError):
            raise NoData

    def process(self, entry, cached):
        yield "members", cached



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from libff.plugin import *


class Mime(Plugin):
    """The "mime" plugin provides information about the content type of files. It uses 'libmagic'
       to guess the mime type and encoding. It requires the 'file-magic' module.
    """

    speed = Speed.SLOW

    use_cache = True
    cache_tag = 0

    attributes = [
        ("mime", String, "The full mime type of the file."),
        ("type", String, "The content type of the file, i.e. the first part of the mime type."),
        ("subtype", String, "The sub type of the file, i.e. the second part of the mime type."),
        ("encoding", String, "The encoding of the file."),
        ("name", String, "The full text description of the type of the file."),
    ]

    @classmethod
    def setup(cls):
        # pylint:disable=global-statement,import-outside-toplevel
        global magic
        try:
            import magic
        except ImportError:
            raise MissingImport("file-magic")

    def can_handle(self, entry):
        return entry.is_file()

    def cache(self, entry):
        # pylint:disable=undefined-variable
        try:
            detected = magic.detect_from_filename(entry.path)
            return detected.mime_type, detected.encoding, detected.name
        except (OSError, ValueError):
            raise NoData

    def process(self, entry, cached):
        mime, encoding, name = cached
        mime_parts = mime.split("/", 1)
        yield "mime", mime
        yield "type", mime_parts[0]
        yield "subtype", mime_parts[1]
        yield "encoding", encoding
        yield "name", name



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from libff.entry import Entry
from libff.plugin import *


class File(Plugin):
    """The "file" plugin provides all essential file attributes. Among these attributes are
       variations of the file's path name (e.g. dirname, basename, extension) and inode information
       from stat(2) (e.g. file size, file type, modification time). The leading 'file.' part is
       optional and can be omitted.
    """

    # This is basically a dummy plugin for the file namespace. Most attributes are redirected
    # directly to the Entry object in Registry.get_attribute().

    # Since 2.5 pylint warns about properties not having __doc__ attributes...
    # pylint:disable=no-member

    speed = Speed.FAST

    attributes = [
        ("path",    Path,     "The full pathname of the file. It will be relative to the current "\
                              "working directory depending on the <directory> arguments that were "\
                              "given on the command line. This can be changed with the "\
                              "-a/--absolute path option."),
        ("root",    Path,     "The start directory the file was found in."),
        ("relpath", Path,     "The pathname of the file relative to the start directory."),
        ("dir",     Path,     "The dirname portion of the file."),
        ("name",    String,   "The basename portion of the file"),
        ("ext",     String,   Entry.ext.__doc__),
        ("pathx",   Path,     Entry.pathx.__doc__),
        ("namex",   Path,     Entry.namex.__doc__),
        ("mode",    Mode,     "The mode and permission bits of the file."),
        ("type",    FileType, "The file type: one of 'd'/'directory', 'f'/'file', 'l'/'symlink', "\
                              "'s'/'socket', 'p'/'pipe'/'fifo', 'char', 'block', 'door', 'port', "\
                              "'whiteout' or 'other'."),
        ("device",  Number,   Entry.device.__doc__),
        ("inode",   Number,   Entry.inode.__doc__),
        ("samedev", Boolean,  Entry.samedev.__doc__),
        ("depth",   Number,   Entry.depth.__doc__),
        ("exec",    Boolean,  Entry.exec.__doc__),
        ("size",    Size,     Entry.size.__doc__),
        ("mtime",   Time,     Entry.mtime.__doc__),
        ("ctime",   Time,     Entry.ctime.__doc__),
        ("atime",   Time,     Entry.atime.__doc__),
        ("time",    Time,     Entry.time.__doc__),
        ("perm",    Mode,     Entry.perm.__doc__),
        ("links",   Number,   Entry.links.__doc__),
        ("uid",     Number,   Entry.uid.__doc__),
        ("gid",     Number,   Entry.gid.__doc__),
        ("user",    String,   Entry.user.__doc__),
        ("group",   String,   Entry.group.__doc__),
        ("hide",    Boolean,  "Whether the the name of the file starts with a dot."),
        ("hidden",  Boolean,  Entry.hidden.__doc__),
        ("empty",   Boolean,  Entry.empty.__doc__),
        ("link",    Path,     "The target path of a symbolic link relative to its parent "\
                              "directory. Empty if the file is not a symbolic link."),
        ("target",  Path,     "The full target path of a symbolic link. Empty if the file is not "\
                              "a symbolic link."),
        ("broken",  Boolean,  "Whether the target of a symbolic link points to a file that does "\
                              "not exist."),
        ("mount",   Boolean,  Entry.mount.__doc__)
    ]

    def can_handle(self, entry):
        return True

    def process(self, entry, cached):
        return {}



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import re
import time
import datetime

UNITS = ("", "K", "M", "G", "T", "P", "E", "Z", "Y")

def format_size(number, base=1024):
    """Convert a number to a string with a human readable file size.
    """
    if number < base:
        return str(number)

    for unit in UNITS:
        if number < base:
            break
        number /= base
        number = round(number, 1)

    if base == 1000:
        unit += "B"

    if number < 10:
        return f"{number:.1f}{unit}"
    else:
        return f"{int(number):d}{unit}"


regex = re.compile(r"(\d+)b?$|(\d+(?:[\.]\d+)?)(" + \
        f"{'|'.join(unit for unit in UNITS if unit)}|" + \
        f"{'|'.join(unit + 'B' for unit in UNITS if unit)}|" + \
        f"{'|'.join(unit + 'iB' for unit in UNITS if unit)}" + \
        ")?$", re.IGNORECASE)

def parse_size(string):
    """Convert a string denoting a file size to an integer.
    """
    string = string.upper()

    match = regex.match(string)
    try:
        number_in_bytes, number, unit = match.groups()
    except AttributeError:
        pass

    if match is None or (number_in_bytes is None and unit is None):
        raise ValueError(f"invalid size {string!r}")

    if number_in_bytes is None:
        prefix = unit[0]
        suffix = unit[1:]

        exp = UNITS.index(prefix)
        base = 1000 if suffix == "B" else 1024
        return int(float(number) * base ** exp)
    else:
        return int(number_in_bytes)


def format_duration(seconds):
    """Format a number of seconds to a human readable duration representation.
    """
    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds %= 60

    if hours:
        return "%dh%dm%ds" % (hours, minutes, seconds)
    else:
        return "%dm%ds" % (minutes, seconds)


regex_duration = re.compile(r"(\d+)([smhdwMy])?")

def parse_duration(string):
    """Convert a human readable duration from a string to seconds.
    """
    string = string.lower()

    try:
        return int(string) * 60
    except ValueError:
        pass

    duration = 0
    end = 0
    for match in regex_duration.finditer(string):
        if match.start(0) != end:
            raise ValueError(f"unable to parse {string!r}")
        end = match.end(1) + 1

        count, unit = match.groups()
        count = int(count)

        if unit == "s":
            duration += count
        elif unit == "m":
            duration += count * 60
        elif unit == "h":
            duration += count * 3600
        elif unit == "d":
            duration += count * 86400
        elif unit == "w":
            duration += count * 604800 # 7 days
        elif unit == "M":
            duration += count * 2592000 # 30 days
        elif unit == "y":
            duration += count * 31536000 # 365 days

    if end != len(string):
        raise ValueError(f"unable to parse {string!r}")

    return duration


def format_time(seconds):
    """Format a number of seconds from epoch to a date and a time.
    """
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(seconds))


time_formats = (
    ("datetime", "%Y-%m-%d %H:%M:%S"),
    ("datetime", "%Y-%m-%d %H:%M"),
    ("datetime", "%Y%m%d%H%M"),
    ("date", "%Y-%m-%d"),
    ("date", "%Y%m%d"),
    ("time", "%H:%M:%S"),
    ("time", "%H:%M"),
    ("time", "%H%M")
)

def parse_time(string):
    """Convert a date and time or a duration to the corresponding number of seconds. A simple
       integer or float is taken as seconds since epoch.
    """
    try:
        return int(float(string))
    except ValueError:
        pass

    if len(string) >= 4:
        for typ, fmt in time_formats:
            try:
                time_obj = datetime.datetime.strptime(string, fmt)
            except ValueError:
                continue
            else:
                if typ == "time":
                    now = datetime.datetime.now()
                    today = datetime.date.today()
                    time_obj = time_obj.replace(year=today.year, month=today.month, day=today.day)
                    if time_obj > now:
                        time_obj -= datetime.timedelta(days=1)
                return int(time_obj.timestamp())

    try:
        return int(time.time() - parse_duration(string))
    except ValueError:
        raise ValueError(f"{string!r} is no valid time")



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import subprocess

from libff.plugin import *


class Filelist(set):
    """The list of filenames that are tracked inside a git repository.
    """

    def __init__(self, dirname):
        super().__init__()

        proc = subprocess.run(["git", "-C", dirname, "ls-files"], capture_output=True, text=True,
                check=False)
        for filename in proc.stdout.splitlines():
            self.add(os.path.join(dirname, filename))


class Status(set):
    """The list of filenames that are tracked inside a git repository and have been changed.
    """

    def __init__(self, dirname):
        super().__init__()

        proc = subprocess.run(["git", "-C", dirname, "status", "--porcelain"], capture_output=True,
                text=True, check=False)
        for line in proc.stdout.splitlines():
            status, filename = line.split(None, 1)
            if status == "??":
                continue
            self.add(os.path.join(dirname, filename))


class Git(Plugin):
    """The "git" plugin provides information about files that are inside a git(1) repository. It
       requires the 'git' executable.
    """

    speed = Speed.SLOW

    attributes = [
        ("tracked", Boolean, "True if the file is tracked by a git repository."),
        ("dirty", Boolean, "True if the file is tracked and has changed or if a directory "\
                "contains a changed file."),
        ("repo_dir", String, "The base directory of the git repository the file or "\
                "directory is in."),
        ("repo", Boolean, "True if the directory contains a git repository.")
    ]

    def __init__(self):
        super().__init__()

        self._cache = {}

    def can_handle(self, entry):
        return True

    def find_repo(self, dirname):
        """Find a .git directory in the current or upper directories and return a list of changed
           files and a list of tracked files.
        """
        if dirname not in self._cache:
            path = os.path.join(dirname, ".git")
            if os.path.isdir(path):
                self._cache[dirname] = dirname, Filelist(dirname), Status(dirname)
            else:
                if dirname == os.sep:
                    self._cache[dirname] = None, None, None
                else:
                    self._cache[dirname] = self.find_repo(os.path.dirname(dirname))

        return self._cache[dirname]

    def process(self, entry, cached):
        dirname = entry.abspath if entry.is_dir() else \
                os.path.dirname(entry.abspath)

        repo_dir, filelist, status = self.find_repo(dirname)

        if repo_dir is None:
            raise NoData

        else:
            data = {"repo_dir": repo_dir, "repo": repo_dir == entry.abspath}

            # Iterate the list of tracked files and the list of changed files and test whether the
            # entry is in one of them. In case of directories it is checked whether it contains a
            # file that is in one of these lists.
            for key, paths in ("tracked", filelist), ("dirty", status):
                if entry.is_dir():
                    abspath = entry.abspath + os.sep
                    data[key] = any(p.startswith(abspath) for p in paths)
                else:
                    data[key] = entry.abspath in paths

            return data



# Put this file in ~/.ff and rename it.

from libff.plugin import *

# IMPORTANT: Do not import third-party modules here, that may not be installed on the system, do
# this in Plugin.setup().


class Foo(Plugin):
    # Give the Plugin a descriptive name that does not yet exist.

    """The "foo" plugin provides access to ...

       (This docstring will appear as the plugin's description in help texts.)
    """

    # Add some information about yourself and how to contact you.
    author = "Jane Doe"
    email = "jane@example.com"
    url = "https://example.com/foo-plugin/"

    # Here you should tell ff(1) roughly how fast this plugin is able to extract data from the each
    # entry. This is important in optimizing test sequences. ff(1) will put faster tests at the
    # beginning and slower tests at the end of a test sequence to reduce the number of evaluations
    # of slower tests. There are three possible settings: FAST, NORMAL (default) and SLOW:
    #
    # FAST:   Plugins that depend only on path names and inode information like stat(2) results or
    #         maybe extended attributes.
    #
    # NORMAL: Use this if you are not sure. Suitable if the plugin opens the file to read the data
    #         but does not do time-consuming processing.
    #
    # SLOW:   If the plugin has to intensively process file data (e.g. decompress it), calls
    #         external programs or libraries, or does other very slow things like web requests etc.
    speed = Speed.NORMAL

    # If use_cache is True it is possible to cache data for each entry that is time-consuming to
    # prepare. The cache() method will be called, its return value will be cached and then passed
    # to process(). Cached values are discarded as soon as the associated entry's modification time
    # changes or the code of this module is altered.
    use_cache = False

    # Explicitly set the cache tag. Rarely used. If cache_tag is None it will be set to the return
    # value of get_plugin_cache_tag(). See there for a more in-depth explanation.
    cache_tag = None

    # Here you tell ff(1) which attributes your plugin will provide. `attributes` is a list of
    # 3-element-tuples: (name, type, description). `type` must be a subclass of Type. A Type's job
    # is for example to describe which operators the value supports, and to format the value for
    # output or to a human readable form. You can create your own Types if needed, see
    # libff/type.py for details.
    #
    # Available Types are:
    #
    #    String             a generic string
    #    ListOfStrings      a list of strings
    #    Number             a number greater or equal to 0
    #    Size               a size in bytes
    #    Time               a point in time in seconds
    #    Duration           a duration in seconds
    #    Boolean            a true or false value
    #
    attributes = [
        ("foo", String, "The name of the file."),
        ("bar", Size, "The size of the file.")
    ]

    @classmethod
    def get_plugin_cache_tag(cls, path):
        # Optional. Is only used if use_cache = True. Rarely used.
        #
        # Return a "tag" that should stay the same as long as the format of the plugin's cached
        # data stays the same. Otherwise, all cached values for this plugin are discarded and
        # recached.
        # The default scheme returns a checksum of the plugins source code, so the cache is
        # invalidated automatically everytime the plugin code is changed.
        # If you need more control over the cache you can implement your own scheme, e.g. return a
        # number that you increment manually everytime you change the data format.

    @classmethod
    def setup(cls):
        # Optional.
        #
        # This classmethod is called once before the Plugin is about to be used. It is supposed to
        # be used for setup code, mainly for importing modules that may or may not be installed.
        #
        #   @classmethod
        #   def setup(cls):
        #       global module
        #       import module

    def can_handle(self, entry):
        # Required.
        #
        # Return True if this plugin can handle this type of file. There are many ways possible for
        # you to do this: based on an attribute of the Entry object (e.g. file type, extension,
        # mime type etc.) or by reading a magic number from the file.
        # This check is supposed to reduce the number that the cache() and process() methods will
        # be called and will also reduce the load on the cache and the number of needlessly cached
        # entries.
        #
        #   def can_handle(self, entry):
        #       return entry.is_file()

    def get_entry_cache_tag(self, entry):
        # Optional. Is only called if use_cache = True. Rarely used.
        #
        # Return a "tag" that indicates the validity of the cached value for this entry. Once it
        # changes the cached value will be discarded and recached. The default method returns the
        # file's modification time which is what you want in most cases.

    def cache(self, entry):
        # Optional. Is only called if use_cache = True.
        #
        # Extract and prepare information from the entry and put it in the cache. After the value
        # has been cached this method will not be called again unless the result of
        # cache_tag(entry) changes. The return value of cache() will be cached and then passed to
        # the process() method. It may have any type. If the file information cannot be extracted,
        # raise a NoData exception.

    def process(self, entry, cached):
        # Required.
        #
        # Extract information from the entry to produce values for the attributes described above.
        # `entry` is a an Entry object from libff/entry.py, see there for attributes and
        # properties. If use_cache is True, `cached` is the cached return value from the cache()
        # method or None if there was a NoData exception raised.
        # On success, either return a dictionary with the values or yield key-value pairs. On
        # failure, i.e. if the file is invalid or the required information cannot be extracted,
        # raise a NoData exception or return an empty dictionary.
        yield "foo", entry.name
        yield "bar", entry.size



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import re
import collections

from .path import join

GroupKey = collections.namedtuple("GroupKey", "include anchored directory")


class Glob:
    """A single gitignore-style pattern.
    """

    def __init__(self, pattern, path_pattern=True):
        self.pattern = pattern

        self.regex_pattern, self.include, self.anchored, self.directory = \
                self.translate(self.pattern, path_pattern)
        self._regex = None

    @property
    def regex(self):
        """The compiled regular expression.
        """
        if self._regex is None:
            self._regex = re.compile(self.regex_pattern)
        return self._regex

    @staticmethod
    def translate(pattern, path_pattern=True):
        """Translate the pattern to a regular expression.
        """
        # pylint:disable=too-many-branches
        if pattern.startswith(r"\#"):
            pattern = pattern[1:]

        # Shall the match be inverted?
        if pattern.startswith("!"):
            include = False
            pattern = pattern[1:]
        else:
            include = True

        if pattern.startswith(r"\!"):
            pattern = pattern[1:]

        # Is there a slash at the beginning or in the middle of the pattern? Then the whole path
        # relative to the directory of the ignore file is matched against.
        if path_pattern:
            try:
                anchored = pattern.index("/") < len(pattern) - 1
            except ValueError:
                anchored = False

            if anchored:
                pattern = pattern.lstrip("/")

        else:
            anchored = False

        # Shall the patten match only directories?
        if pattern.endswith("/"):
            directory = True
            pattern = pattern[:-1]
        else:
            directory = False

        # Translate the wildcards to regular expression syntax.
        parts = []
        for part in re.split(r"(/\*\*/|\*\*/|/\*\*|\*|\?|\[![^\]]+\]|\[[^\]]+\])", pattern):
            if part == "/**/":
                part = "(?:/|/.+/)"
            elif part == "**/":
                part = "(?:.+/)?"
            elif part == "/**":
                part = "(?:/.+)?"
            elif part == "*":
                part = "[^/]*"
            elif part == "?":
                part = "[^/]"
            elif part.startswith("[!"):
                part = f"[^{part[2:-1]}]"
            elif part.startswith("["):
                part = f"[{part[1:-1]}]"
            else:
                part = re.escape(part)
            parts.append(part)

        return "^" + "".join(parts) + "$", include, anchored, directory

    def match(self, path, basename=None, is_dir=None):
        """Match a path against the pattern. If the path matches the pattern, you must use the
           .include attribute to determine whether the path is to be added or removed from the
           result. path must be relative. basename and is_dir can be used to provide pre-computed
           values.
        """
        if basename is None:
            basename = os.path.basename(path)
        if is_dir is None:
            is_dir = os.path.isdir(path)

        # Turn path into a pseudo-relative path.
        path = path.lstrip(os.sep)

        if self.regex.match(path if self.anchored else basename):
            if self.directory:
                if is_dir:
                    return True
            else:
                return True
        return False

    def get_group_key(self):
        """Return an object that can be used as a key for grouping different types of glob
           patterns.
        """
        return GroupKey(self.include, self.anchored, self.directory)


class GitIgnore:
    """Parse a .gitignore file according to the rules in the gitignore(5) manpage.
    """
    # XXX See gitignore(5). This class does not fully support all features of gitignore described
    # in the manpage yet. It does not read $GIT_DIR/info/exclude or read core.excludesFile from the
    # git(1) configuration.

    IGNORE_NAMES = set([".gitignore", ".ignore", ".fdignore", ".ffignore"])

    @classmethod
    def find_ignore_files(cls, dirname):
        """Find ignore files in the current and in the parent directories of 'dirname', that must
           be applied.
        """
        assert os.path.isabs(dirname)
        ignores = []
        parent = os.sep
        for part in dirname.split(os.sep):
            parent = join(parent, part)
            for name in cls.IGNORE_NAMES:
                if os.path.exists(join(parent, name)):
                    ignores.append((parent, name))
        return ignores

    @classmethod
    def from_parent_directories(cls, context, path):
        """Return a list of GitIgnore objects from the parent directories of 'path'.
        """
        path = os.path.abspath(path)
        ignores = []
        for parent, name in cls.find_ignore_files(os.path.dirname(path)):
            try:
                ignores.append(cls(parent, name))
            except OSError as exc:
                context.logger.warning(exc)
        return ignores

    def __init__(self, dirname, name):
        assert os.path.isabs(dirname)
        self.dirname = dirname
        self.path = join(dirname, name)

        with open(self.path) as lines:
            self.lines = list(lines)

        # Collect the glob patterns from the gitignore file and translate them to regular
        # expressions. We have to evaluate the patterns from the file in sequence, so in order to
        # speed up matching we compile them into groups of similar patterns which we can join into
        # one big regex. Patterns differ in whether they are inclusive or exclusive, aimed at full
        # paths and whether they are supposed to match directories only.
        self.patterns = []
        key = None
        for line in self.lines:
            pattern = line.strip()
            if not pattern or pattern.startswith("#"):
                continue

            pattern = Glob(pattern)
            if pattern.get_group_key() != key:
                # Start new group.
                key = pattern.get_group_key()
                patterns = []
                self.patterns.append((key, patterns))
            patterns.append(pattern)

        self.patterns = [
                (key, re.compile("|".join(pattern.regex_pattern for pattern in patterns)))
                for key, patterns in self.patterns]

    @staticmethod
    def match(ignores, path, name, is_dir):
        """Match the path or basename against the sequence of patterns from all the ignore files
           and return True if the entry is to be excluded. Also return the path to the ignore file
           that contained the matching pattern.
        """
        assert os.path.isabs(path)

        ignore_path = None
        exclude = False
        for ignore in ignores:
            relpath = path[len(ignore.dirname) + 1:]

            for key, pattern in ignore.patterns:
                if key.directory and not is_dir:
                    continue
                test_path = relpath if key.anchored else name
                if pattern.match(test_path):
                    ignore_path = ignore.path
                    exclude = key.include

        return exclude, ignore_path if exclude else None



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import ast

from libff.plugin import *


class Elf(Plugin):
    """The "elf" plugin provides information extracted from ELF executable files. It requires the
       'pyelftools' module.
    """

    use_cache = True

    attributes = [
        ("sonames", ListOfStrings, "The names of all shared objects that are linked in this "\
                                   "executable file.")
    ]

    def setup(self):
        # pylint:disable=global-statement,import-outside-toplevel,global-variable-not-assigned
        global ELFFile, DynamicSection, ELFError
        try:
            from elftools.elf.elffile import ELFFile
            from elftools.elf.dynamic import DynamicSection
            from elftools.common.exceptions import ELFError
        except ImportError:
            raise MissingImport("pyelftools")

    def extract_sonames(self, path):
        """Yield the names of the shared objects required by an ELF executable.
        """
        # pylint:disable=undefined-variable
        try:
            with open(path, "rb") as fobj:
                elffile = ELFFile(fobj)
                for section in elffile.iter_sections():
                    if not isinstance(section, DynamicSection):
                        continue

                    for tag in section.iter_tags():
                        if tag.entry.d_tag == "DT_NEEDED":
                            yield tag.needed
        except (OSError, ELFError):
            raise NoData

    def can_handle(self, entry):
        if entry.is_file():
            try:
                with open(entry.path, "rb") as fobj:
                    return fobj.read(4) == b"\x7fELF"
            except OSError:
                pass
        return False

    def cache(self, entry):
        return sorted(self.extract_sonames(entry.path))

    def process(self, entry, cached):
        yield "sonames", cached


class Shebang(Plugin):
    """The "shebang" plugin extracts the shebang line from a script, i.e. the first line of the
       file if it starts with '#!'.
    """

    attributes = [
        ("shebang", String, "The contents of the shebang line (#!).")
    ]

    @staticmethod
    def extract_shebang(path):
        """Extract the shebang line.
        """
        try:
            with open(path, "rb") as lines:
                for line in lines:
                    if line.startswith(b"#!"):
                        return line[2:].decode("ascii").rstrip()
        except (OSError, UnicodeDecodeError):
            pass

        return None

    def can_handle(self, entry):
        return entry.is_file()

    def process(self, entry, cached):
        shebang = self.extract_shebang(entry.path)
        if shebang is not None:
            yield "shebang", shebang


class Py(Plugin):
    """The "py" plugin provides information about Python scripts.
    """

    attributes = [
        ("imports", ListOfStrings, "A list of module and package names that are imported "\
                                   "in a Python file.")
    ]

    def parse_imports(self, entry):
        """Extract the names of modules and packages that a Python script imports.
        """
        with open(entry.path) as fobj:
            data = fobj.read()

        for node in ast.walk(ast.parse(data)):
            if isinstance(node, ast.ImportFrom):
                if node.module is not None:
                    if node.level > 0:
                        parts = [node.module]
                        dirname = entry.dir
                        for _ in range(node.level):
                            dirname, basename = os.path.split(dirname)
                            parts.insert(0, basename)
                        yield ".".join(parts)
                    else:
                        yield node.module

            elif isinstance(node, ast.Import):
                yield node.names[0].name

    def can_handle(self, entry):
        if entry.ext.lower() == "py":
            return True
        shebang = Shebang.extract_shebang(entry.path)
        return shebang is not None and "python" in shebang

    def process(self, entry, cached):
        try:
            yield "imports", sorted(self.parse_imports(entry))
        except (OSError, SyntaxError, UnicodeDecodeError):
            raise NoData



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import re
import collections

from . import BaseClass
from .type import Path, Type
from .entry import Entry
from .ignore import Glob
from .attribute import Attribute
from .exceptions import UsageError, ExpressionError


class ParserError(Exception):
    """Exception used by the Parser.
    """


class AND(list):
    """A list of tests that must all be true.
    """
    name = "AND"

    def __init__(self, obj=None):
        super().__init__()
        if obj is not None:
            self.append(obj)

    def __str__(self):
        return f"{self.name}( {', '.join(str(x) for x in self)} )"

class OR(AND):
    """A list of tests either of which must be true.
    """
    name = "OR"

class NOT(AND):
    """A list of tests with its result inverted.
    """
    name = "NOT"


class Test(collections.namedtuple("Test", "attribute operator type value ignore_case")):
    """Test object that contains all relevant information about a test.
    """

    name = "test"

    def format(self, with_type=False):
        """Return a readable representation of the Test.
        """
        if isinstance(self.value, (re.Pattern, Glob)):
            value = self.value.pattern
        else:
            value = self.value

        if with_type:
            return f"{self.attribute.fullname}[{self.type.name}]{self.operator}{value}"
        else:
            return f"{self.attribute.fullname}{self.operator}{value}"

    def __str__(self):
        return self.format(with_type=True)

    def __repr__(self):
        return repr(str(self))


class FlatParser(BaseClass):
    """Parse a list of test expressions into a test sequence.
    """

    expression_regex = re.compile(r"^\s*((?:\w+\.)?\w+?)\s*"\
                                  r"(=|:|~|%|>=|>|<=|<|\+=|\+|\-=|\-)"\
                                  r"(\{(?:\w+\.)?\w+?\}|\{\})?"\
                                  r"\s*(.+)\s*$")

    def __init__(self, context, tokens):
        super().__init__(context)

        self.tokens = tokens.copy()
        self.sequence = self.parse()

        self.optimize(self.sequence)

    def get_speed_score(self, test):
        """Return the speed score from a single test or the sum of scores from all tests of a sub
           sequence.
        """
        if isinstance(test, Test):
            return self.registry.get_plugin_speed(test.attribute)
        else:
            return sum(self.get_speed_score(t) for t in test)

    def optimize(self, sequence):
        """Optimize a test sequence, i.e. sort all sequences and their tests according to speed
           score, so that faster tests are evaluated first.
        """
        sequence.sort(key=self.get_speed_score)

        for seq in sequence:
            if isinstance(seq, list):
                self.optimize(seq)

    def __iter__(self):
        yield from self.sequence

    def parse(self):
        """Parse a list of tokens into a test structure.
        """
        sequences = AND()
        self.parse_sequence(sequences)
        return sequences

    def format(self, tests=None, level=0):
        """Format a list of tests for output.
        """
        if tests is None:
            tests = self.sequence

        yield ("    " *  level) + tests.name + "("

        for test in tests:
            if isinstance(test, list):
                yield from self.format(test, level + 1)
            else:
                yield ("    " * (level + 1)) + str(test)

        yield ("    " *  level) + ")"

    def parse_test(self, test):
        """Parse an expression and return a Test object.
        """
        match = self.expression_regex.match(test)
        if match is not None:
            attribute, operator, reference, value = match.groups()
            attribute = self.registry.setup_attribute(attribute)
            if reference is not None:
                reference = reference.strip("{}")

            # Normalize operators that contain > and <.
            operator = operator.replace(">", "+").replace("<", "-")

            if attribute.plugin is None:
                attribute = Attribute("file", attribute.name)

            return self.create_test(attribute, operator, reference, value)

        else:
            raise ExpressionError(f"Invalid expression {test!r}!")

    def get_reference_value(self, type_cls, attribute, reference, value):
        """Fetch the value from the reference.
        """
        if reference:
            ref_attribute = self.registry.setup_attribute(reference)
        else:
            ref_attribute = attribute

        ref_type_cls = self.registry.get_attribute_type(ref_attribute)

        # Check if both attributes are comparable, i.e. they have a similar type.
        try:
            self.get_common_super_type(type_cls, ref_type_cls)
        except ValueError:
            raise ExpressionError(f"{attribute} and {ref_attribute} have different types "\
                    "and cannot be compared")

        entry = Entry.as_reference(self.args, value)
        return self.registry.get_attribute(entry, ref_attribute)

    @staticmethod
    def get_common_super_type(a_type, b_type):
        """Find the common Type subclass of both Type subclasses. If both arguments have a common
           Type, i.e. Number or String, they are comparable.
        """
        assert issubclass(a_type, Type)
        assert issubclass(b_type, Type)

        for cls in a_type.mro():
            if cls is Type:
                break
            elif cls in b_type.mro():
                return cls
        raise ValueError("no common Type subclass found")

    def create_test(self, attribute, operator, reference, value):
        """Create the Test object. Before that, do some sanity checks, handle case sensitivity,
           prepare regular expressions and fetch the value from a reference file if required.
        """
        # pylint:disable=broad-except,too-many-branches
        type_cls = self.registry.get_attribute_type(attribute)

        if operator not in type_cls.operators:
            raise ExpressionError(f"Attribute {attribute} of type {type_cls.name!r} does not "\
                    f"support operator {operator!r}")

        ignore_case = None

        try:
            # We either use the provided value directly or in case of a reference we use the value
            # from after evaluating the reference.
            if reference is not None:
                value = self.get_reference_value(type_cls, attribute, reference, value)
            else:
                value = type_cls.input(value)

        except Exception as exc:
            raise UsageError(str(exc))

        if type_cls.choices is not None and value not in type_cls.choices:
            raise ExpressionError(f"You specified an invalid value {value!r} "\
                    f"for attribute '{attribute}'! Allowed values are: " + \
                    ",".join(sorted(type_cls.choices)) + ".")

        # Do some preparation for Type classes that claim to be string types, i.e. handle case
        # sensitivity and compile regexes.
        if type_cls.string_type:
            if self.args.case == "smart":
                # If the value is in lower case compare ignoring the case, if it contains uppercase
                # letters compare using sensitive case.
                ignore_case = value == value.lower()
            else:
                ignore_case = self.args.case == "ignore"

            if ignore_case:
                value = value.lower()

            # Pre-compile regular expressions.
            if operator == "~":
                try:
                    value = re.compile(value)
                except re.error as exc:
                    raise ExpressionError(f"Invalid regex pattern {value!r}: {exc}")

            elif operator == "%":
                value = Glob(value, path_pattern=issubclass(type_cls, Path))

        return Test(attribute, operator, type_cls, value, ignore_case)

    def parse_sequence(self, sequences):
        """Parse a part of a list of tokens into a sequence of tests.
        """
        while self.tokens:
            token = self.tokens.pop(0)
            sequences.append(self.parse_test(token))


class Parser(FlatParser):
    """Parse a list of test expressions and operators into a nested test structure.
    """

    OPENING_BRACKETS = ("(", "{{")
    CLOSING_BRACKETS = (")", "}}")

    def parse(self):
        """Parse a list of tokens into a test structure.
        """
        sequences = OR(AND())
        self.parse_sequence(sequences)
        if self.tokens:
            raise ParserError(f"superfluous closing bracket {self.tokens[0]!r}")
        return sequences

    def parse_bracket_sequence(self, sequences, sub_sequences):
        """Parse a part of a list of tokens inside brackets.
        """
        self.parse_sequence(sub_sequences)

        if not self.tokens:
            raise ParserError("incomplete sub sequence")
        self.tokens.pop(0)

        sequences[-1].append(sub_sequences)

    def parse_sequence(self, sequences):
        """Parse a part of a list of tokens into a sequence of tests.
        """
        # pylint:disable=too-many-branches

        while self.tokens:
            token = self.tokens.pop(0)
            utoken = token.upper()

            if token in self.OPENING_BRACKETS:
                self.parse_bracket_sequence(sequences, OR(AND()))
                continue

            elif token in self.CLOSING_BRACKETS:
                if not sequences[-1]:
                    raise ParserError("empty expression")
                self.tokens.insert(0, token)
                return

            elif utoken == "AND":
                continue

            elif utoken == "OR":
                sequences.append(AND())
                continue

            elif utoken == "NOT":
                if not self.tokens:
                    raise ParserError("premature eof")

                utoken = self.tokens[0].upper()

                if utoken in self.OPENING_BRACKETS:
                    self.tokens.pop(0)
                    self.parse_bracket_sequence(sequences, NOT(AND()))

                elif utoken in ("AND", "OR", "NOT") + self.CLOSING_BRACKETS:
                    raise ParserError(f"unexpected token {utoken!r}")

                else:
                    sequences[-1].append(NOT(AND(self.parse_test(self.tokens.pop(0)))))

                continue

            sequences[-1].append(self.parse_test(token))

        if len(sequences) > 1 and any(not seq for seq in sequences):
            raise ParserError("empty expression")



#!/usr/bin/python3
# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import sys
import glob
from distutils.sysconfig import get_python_lib

import setuptools

from libff.__version__ import __version__

with open("README.md") as fobj:
    long_description = fobj.read()

kwargs = {
    "name":         "find-ff",
    "version":      str(__version__),
    "author":       "Lars Gustäbel",
    "author_email": "lars@gustaebel.de",
    "url":          "https://github.com/gustaebel/ff/",
    "description":  "A tool for finding files in the filesystem",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "license":      "GPLv3+",
    "classifiers":  ["Development Status :: 3 - Alpha",
                     "Environment :: Console",
                     "Intended Audience :: Developers",
                     "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
                     "Operating System :: POSIX",
                     "Operating System :: POSIX :: Linux",
                     "Topic :: System :: Filesystems",
                     "Topic :: Utilities",
                     "Programming Language :: Python :: 3"],
    "python_requires": ">=3.6",
    "install_requires": ["pymediainfo", "file-magic"],

    "packages":     ["ff", "libff", "libff.builtin"],
    "scripts":      ["bin/ff"],
    "data_files":   [("/usr/share/man/man1", ["man/ff.1"]),
                     ("/usr/share/man/man7", ["man/ff.7"]),
                     ("/usr/lib/ff", glob.glob("plugins/*.py")),
                     ("/usr/share/ff", ["plugin_template.py"]),
                     (os.path.join(get_python_lib(), "libff"), ["libff/manpage.template"])]
}

setuptools.setup(**kwargs)



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import sys
import textwrap

from . import OUTPUT_WIDTH


class Logger:
    """Class that centralizes logging.
    """

    tags_emitted = set()

    def __init__(self):
        self.debug_categories = None

    def set_debug(self, debug_categories):
        """Set the list of debug categories to display.
        """
        self.debug_categories = debug_categories

    def print_message(self, category, message, tag=None):
        """Format, wrap and print the text of a message for output.
        """
        if tag is not None:
            if tag in self.tags_emitted:
                return
            else:
                self.tags_emitted.add(tag)

        print(textwrap.fill(category + ": " + str(message), width=OUTPUT_WIDTH,
                subsequent_indent=" " * (len(category) + 2)), file=sys.stderr)

    def warning(self, message, tag=None):
        """Print a warning to standard error output. If tag is not None hide the warning if there
           has already been another warning with that tag.
        """
        self.print_message("WARNING", message, tag)

    def hint(self, message, tag=None):
        """Print a hint to standard error output. A hint is some kind of "soft" warning.
        """
        self.print_message("HINT", message, tag)

    def error(self, message, exitcode):
        """Print an error message to standard error output and exit.
        """
        self.print_message("ERROR", message)
        if exitcode is not None:
            raise SystemExit(exitcode)

    def info(self, message):
        """Print an info message to standard error output.
        """
        self.print_message("INFO", message)

    if __debug__:
        def debug(self, category, message):
            """Print a debug message to standard error output if --debug is switched on.
            """
            if "all" in self.debug_categories or category in self.debug_categories:
                print(f"DEBUG:{category}: {message}", file=sys.stderr)

        def debug_proc(self, index, message):
            """Print a debug message to standard error output if --debug-processes is switched on.
            """
            self.debug("mp", f"process #{index:02d} {message}")

    def exception(self, message, traceback, exitcode=None):
        """Print an error message and a traceback to standard error output and exit.
        """
        self.print_message("INTERNAL", message + ":")
        sys.stderr.write(traceback)
        if exitcode is not None:
            raise SystemExit(exitcode)



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import io
import os
import re
import datetime
import subprocess

from . import exceptions
from .convert import time_formats
from .arguments import create_parser
from .__version__ import __version__

bd = r"\fB"
it = r"\fI"
rs = r"\fR"
br = ".br"


class ManPage:
    """Programmatically create a manpage from text-format parts. The format is similar to that of
       txt2man(1).
    """

    name = "ff"
    title = "ff - Find files in the filesystem"
    section = 1
    description = None

    def __init__(self):
        self.lines = []
        self.render()

    def render(self):
        """Render the manpage.
        """
        self.start()

    def wrap(self, lines):
        """Process a list of lines and wrap paragraphs into single lines.
        """
        current_level = 0
        text = ""
        for line in lines:
            line = line.rstrip()

            level = len(line) - len(line.lstrip())

            if level < current_level or (current_level == 0 and level > 0) or "  " in line.lstrip():
                # Start a new paragraph if the line is empty, its indentation level is smaller than
                # before, if the previous line was a section header line, or the line contains the
                # start of a definition (identified by two consecutive spaces like txt2man(1)'s tag
                # list).
                if text:
                    yield text
                yield ""
                text = line

            else:
                # The line is a continuation line, append it to the current paragraph.
                if not text:
                    text = line
                else:
                    text += " " + line.lstrip()

            if line:
                current_level = level

        if text:
            yield text

    def include(self, name):
        """Replace an include statement.
        """
        if name == "help":
            fobj = io.StringIO()
            parser = create_parser()
            parser.print_manpage_help(file=fobj)
            self.parse_lines(fobj.getvalue().splitlines())

        elif name == "usage":
            fobj = io.StringIO()
            parser = create_parser()
            parser.print_usage(file=fobj)
            self.parse_lines(fobj.getvalue().splitlines())

        elif name == "time_patterns":
            lines = []
            for _, fmt in time_formats:
                lines.append("    ``" + re.sub(r"%([YmdHMS])", lambda m: m.group(1)*2, fmt) + "``")
                lines.append("")
            self.parse_lines(lines)

        elif name == "exit_codes":
            lines = []
            for obj in vars(exceptions).values():
                # pylint:disable=unidiomatic-typecheck
                if type(obj) is type and issubclass(obj, exceptions.BaseError) and \
                        obj is not exceptions.BaseError:
                    lines.append(f"    {obj.exitcode}  {obj.__doc__}")
            self.parse_lines(lines)

        else:
            raise ValueError(f"invalid include statement {name!r}")

    def parse_lines(self, lines):
        """Parse a set of lines from a manpage template.
        """
        lines = self.wrap(lines)

        for line in lines:
            line = line.rstrip()

            if line.lstrip().startswith("::"):
                # Evaluate an include marker.
                name = line.lstrip()[2:]
                self.include(name)

            elif re.match(r"^[A-Z ]+$", line):
                # Identify section headers (all caps starting at column 0).
                self.add_section(line)

            elif line.lstrip().startswith("$"):
                # Highlight an example shell command.
                self.lines.append(bd + line + rs)

            elif "  " in line.strip():
                # Format a definition, i.e. a word described with a text.
                key, value = line.strip().split("  ", 1)
                self.add_definition(key)
                self.add(value)
                self.lines.append(".PP")

            else:
                self.add(line)

    def process(self, string):
        """Process a string containing one of more lines and produce roff(7) output.
        """
        # Turn paragraphs into one long line each.
        string = "\n\n".join(" ".join(s.split()) for s in string.split("\n\n"))

        def replace(match):
            if match.group(1):
                # Make single-quoted strings italic. It should match "'s'" but not "file's".
                return f"'{it}{match.group(1)}{rs}'"
            elif match.group(2):
                # Make double backtick-quoted strings italic.
                return f"{it}{match.group(2)}{rs}"
            elif match.group(3):
                # Make backtick-quoted strings bold.
                return f"{bd}{match.group(3)}{rs}"
            elif match.group(4):
                # Highlight references to other manpages.
                return f"{bd}{match.group(4)}{rs}{match.group(5)}"
            else:
                # Highlight short and long options.
                return f"{bd}{match.group(6)}{rs}"

        string = re.sub(r"'(?!s\s)([^']+)'|"\
                        r"``([^`]+)``|"\
                        r"`([^`]+)`|"\
                        r"([a-z_-]+)(\(\d\))|"\
                        r"(?<![a-zA-Z])(\-[a-zA-Z]|\-\-[a-zA-Z-]+)", replace, string)

        # Escape dashes.
        return string.replace("-", "\\-")

    def add(self, line=""):
        """Add a line to the internal buffer.
        """
        self.lines.append(self.process(line))

    def add_header(self, name, section, title):
        """Add a manpage header.
        """
        self.lines.append(f".TH {name} {str(section)} \"{datetime.date.today()}\" "\
                    f"\"Version {__version__}\" \"{title}\"")
        self.lines.append(".nh") # turn off hyphenation

    def add_section(self, name):
        """Add a section.
        """
        self.lines.append(f".SH {name.upper()}")

    def add_subsection(self, name):
        """Add a subsection.
        """
        self.lines.append(f".SS {name}")

    def add_definition(self, name):
        """Add a definition.
        """
        self.lines.append(f".IP \"{bd}{name}{rs}\" 4")

    def start(self):
        """Initialize the manpage.
        """
        self.add_header(self.name, self.section, self.title)
        if self.description:
            self.add_section("Description")
            self.add(self.description)
            self.add()

    def render_attributes(self, plugin):
        """Render a list of attributes of a plugin.
        """
        for attr, type_cls, help_text in plugin.attributes:
            if plugin.name == "file":
                attribute = f"[file.]{attr}"
            else:
                attribute = f"{plugin.name}.{attr}"

            self.add_definition(attribute)
            self.add(f"Type: {type_cls.name}  ( {' '.join(type_cls.operators)} )")
            self.add(br)
            self.add(help_text)
            self.add()

        self.add()

    def render_plugin_detail(self, plugin, full=False):
        """Render the details of a specific plugin with or without author information.
        """
        names = ("name", "source", "path", "use_cache")
        if full:
            names += ("author", "email", "url")

        for name in names:
            if name == "use_cache":
                self.add(f"Uses cache: {'Yes' if plugin.use_cache else 'No'}")
            else:
                self.add(f"{name.capitalize()}: {getattr(plugin, name)}")
            self.add(br)

    def print(self):
        """Print the raw manpage to stdout.
        """
        print("\n".join(self.lines))

    def show(self):
        """Show the manpage using man(1) reading from stdin.
        """
        try:
            subprocess.run(["man", "-l", "-"], input="\n".join(self.lines), text=True, check=True)
        except FileNotFoundError:
            raise SystemExit("unable to call man(1), is it installed?")
        except subprocess.CalledProcessError:
            raise SystemExit("calling man(1) failed")


class FullManPage(ManPage):
    """Create the ff(1) manpage.
    """

    def render(self):
        super().render()

        with open(os.path.join(os.path.dirname(__file__), "manpage.template")) as lines:
            self.parse_lines(lines)


class AttributesManPage(ManPage):
    """Create the ff(7) manpage.
    """

    title = "ff Plugin Attributes Reference"
    section = 7
    description = """
        This is a list of all the attributes that are available. Attributes are provided by
        plugins. For more details on a specific plugin use 'ff --help <plugin>'. Please note that
        if you got this help text by using 'man 7 ff', the information you get is limited to the
        builtin plugins. Use 'ff --help-attributes' to get the full attribute list from all plugins
        that are available right now.
    """

    def __init__(self, plugins):
        self.plugins = plugins
        super().__init__()

    def render_plugin(self, plugin):
        """Render a plugin section.
        """
        self.add_section(f"{plugin.name} Plugin")
        self.add()
        self.add(plugin.__doc__)
        self.add()
        self.render_attributes(plugin)

    def render(self):
        super().render()

        for plugin in self.plugins:
            self.render_plugin(plugin)

        self.add_section("See Also")
        self.add("ff(1)")


class PluginManPage(ManPage):
    """Create a plugin manpage/helptext.
    """

    def __init__(self, plugin):
        self.plugin = plugin

        self.name = f"ff-{plugin.name}"
        self.title = f"{plugin.name} Plugin Reference"
        self.description = plugin.__doc__

        super().__init__()

    def render_plugin(self, plugin):
        """Render a plugin section.
        """
        self.add_section("Attributes")
        self.add()
        self.render_attributes(plugin)

    def render(self):
        super().render()

        self.render_plugin(self.plugin)
        self.add_section("Details")
        self.render_plugin_detail(self.plugin, full=True)

        self.add_section("See Also")
        self.add("ff(1), ff(7)")


class TypesManPage(ManPage):
    """Create manpage/helptext about types.
    """

    name = "ff-types"
    title = "ff Types Reference"
    section = 7
    description = """
        This is a list of all the types that are available. Furthermore, there is information about
        which test operators are supported by each type and if and how attributes of certain types
        will be counted ('ff --count').
    """

    def __init__(self, types):
        self.types = types
        super().__init__()

    def render_type(self, type_cls):
        """Render a type section.
        """
        self.add_section(f"{type_cls.name} Type")
        self.add(f"Name:  {type_cls.name}")
        self.add(br)
        self.add(f"Operators:  {' '.join(type_cls.operators)}")
        self.add(br)
        self.add(f"Count: {type_cls.count.name.lower()}")
        self.add()
        self.add(type_cls.__doc__)
        self.add()

    def render(self):
        super().render()

        for type_cls in self.types:
            self.render_type(type_cls)

        self.add_section("See Also")
        self.add("ff(1), ff(7)")


class PluginsManPage(ManPage):
    """Create manpage/helptext about all available plugins.
    """

    name = "ff-plugins"
    title = "ff Plugins Reference"
    section = 7
    description = """
        This is a list of all the plugins that are currently available. For more details on a
        specific plugin use 'ff --help <plugin>'.
    """

    def __init__(self, plugins):
        self.plugins = plugins
        super().__init__()

    def render_plugin(self, plugin):
        """Render a plugin section.
        """
        self.add_section(f"{plugin.name} Plugin")
        self.render_plugin_detail(plugin)
        self.add()
        self.add(plugin.__doc__)

    def render(self):
        super().render()

        for plugin in self.plugins:
            self.render_plugin(plugin)

        self.add_section("See Also")
        self.add("ff(1), ff(7)")



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import queue
import threading
import multiprocessing
import multiprocessing.sharedctypes

from . import TIMEOUT


class Context:
    """The Context class contains a global state and is globally accessible so that all components
       find all other components.
    """
    # pylint:disable=too-many-instance-attributes

    warnings_emitted = set()

    def __init__(self):
        # These are added later.
        self.args = None
        self.logger = None
        self.cache = None
        self.registry = None
        self.matcher = None
        self.excluder = None
        self.walker = None
        self.processing = None

        self.global_lock = None
        self.stop_queue = None
        self.barrier = None
        self.exitcode_object = None

        self.cache_hits = None
        self.cache_misses = None

    def setup(self):
        """Set up multiprocessing and inter-process communication.
        """
        self.global_lock = multiprocessing.Lock()

        # We use a Queue object to emulate a stop event, because multiprocessing.Event did not work
        # the way I expected it.
        self.stop_queue = multiprocessing.Queue(self.args.jobs)

        # Create a Barrier with the number of processes plus one for the main process.
        if __debug__ and self.args.profile:
            num_jobs = 1
        else:
            num_jobs = self.args.jobs + 1
        self.barrier = multiprocessing.Barrier(num_jobs)

        # The exitcode object provides a way for --exec and --exec-batch processes to feed back
        # errors to the main process.
        self.exitcode_object = multiprocessing.sharedctypes.RawValue("i", 0)

        # Provide a shared counter for cache hits and misses.
        self.cache_hits = multiprocessing.Value("L")
        self.cache_misses = multiprocessing.Value("L")

    def close(self):
        """Close the Context and clean up.
        """
        self.stop_queue.close()

    def set_exitcode(self, exitcode):
        """Set the shared resource 'exitcode' to the value that will be used as the main exit code.
        """
        self.exitcode_object.value = exitcode

    @property
    def exitcode(self):
        """The exitcode to use as the main exit code.
        """
        return self.exitcode_object.value

    def barrier_wait(self, timeout=TIMEOUT):
        """Wait for all processes to reach the Barrier. If all processes do that means that there
           is no more input and processing is finished.
        """
        try:
            self.barrier.wait(timeout)
        except threading.BrokenBarrierError:
            self.barrier.reset()
            return self.is_stopping()
        else:
            return True

    def idle_processes(self):
        """Return the number of idle processes, i.e. processes that are waiting for the barrier
           because there is nothing in the queue.
        """
        return self.barrier.n_waiting

    def is_stopping(self):
        """Return True all Processes are supposed to stop.
        """
        try:
            self.stop_queue.get_nowait()
        except queue.Empty:
            return False
        else:
            return True

    def stop(self):
        """Stop all processes immediately.
        """
        for _ in range(self.args.jobs):
            self.stop_queue.put(None)
